// C-File for pressure modul
// Programmer: Dr. Ralf Wieland
// Version: 0.1 17.11.2002
// Version: 0.2 15.01.2003
// Version: 0.3 01.06.2003 by Andreas Theofilu
// Version: 0.4 07.06.2003 by Andreas Theofilu

#include "druck.h"
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/param.h>
#include "systype.h"
#include "stderror.h"

/*
 * Some global variables we need in this module.
 */
char *DataFile;
int Datafd;

typedef struct
{
	int pressure;
	int temperature;
	time_t tstamp;
} mpoint;

/*
 * Prototypes of all functions in this module.
 */
void SendeBit(int hi);
int LeseBit();
void ResetSensor(void);
unsigned int AdcLesen();
unsigned int CalDatenLesen(unsigned befehl);
unsigned int ADWertLesen(unsigned befehl);
int ParameterLesen();
int DruckTempLesen();
void init_ring();
void insert_ring(int val);
void *pthr_getpressure( void *pV_data );
void *pthr_sendpressure( void *pV_data );
void daemon_start (int ignsigcld);
void sig_child ();

/**************************************************************************/
/********************** Interface to the sensor ***************************/
/**************************************************************************/

void SendeBit(int hi)
{
    unsigned char q;
    if (hi)
	q = 0xfb;
    else
	q = 0xff;
    outb(q, m_iPort);
    usleep(25);
    outb(q & 0xfd, m_iPort);
    usleep(25);
    outb(q, m_iPort);
}

// Bit auslesen
int LeseBit()
{
  return(!(inb(m_iPort + 1) & 32));
}


// Reset des Drucksensors ausfhren
void ResetSensor(void)
{
    int i;
    for(i = 0; i < 8; i++)
    {
	SendeBit(1);
	SendeBit(0);
    }
    for(i = 0; i < 5; i++)
	SendeBit(0);
}


// AD-Wandler oder Cal-Daten auslesen
unsigned int AdcLesen()
{
    int i;
    unsigned int wert;
    SendeBit(0);
    // Wandler starten
    // warten bis Wandlung beendet
    for(i=0; i<25; i++){
	usleep(1);
	if(!LeseBit()) break;
    }
        
    // 16 Bit auslesen
    wert = 0;
    for(i = 0; i < 16; i++)
    {
	SendeBit(0);
	wert <<= 1;
	if (LeseBit())
	    wert |= 1;
    }
    SendeBit(0);
    return(wert);
}



// Auslesen eines Kalibrierwortes
unsigned int CalDatenLesen(unsigned befehl)
{
    int i;
    // Befehl senden  
    for (i = 0; i < 12; i++)
    {
	SendeBit(befehl & 1);
	befehl >>= 1;
    }
    // Wert einlesen
    return(AdcLesen());
}



// Auslesen eines AD-Wertes
unsigned int ADWertLesen(unsigned befehl)
{
    int i;
    // Befehl senden  
    for (i = 0; i < 10; i++)
    {
	SendeBit(befehl & 1);
	befehl >>= 1;
    }
    SendeBit(0);
    // Wandler auslesen
    return(AdcLesen());
}

// Auslesen der Kalibrierdaten des Drucksensors
// Speichern der Daten im m_wC1 .. m_wC6


int ParameterLesen()
{
    unsigned int w1,w2,w3,w4;
    
    if (!m_iPort)
	return(false);
    ResetSensor();
    ResetSensor();
    
    w1 = CalDatenLesen(0x1157);		// Wort 1 lesen
    w2 = CalDatenLesen(0x00d7);		// Wort 2 lesen
    w3 = CalDatenLesen(0x1137);		// Wort 3 lesen
    w4 = CalDatenLesen(0x00b7);		// Wort 4 lesen
    
    // Parameter aufbereiten
    parameterf.m_wC1 = w1 >> 1;
    parameterf.m_wC2 = ((w3 & 0x3f) << 6) | (w4 & 0x3f);
    parameterf.m_wC3 = w4 >> 6;
    parameterf.m_wC4 = w3 >> 6;
    parameterf.m_wC5 = ((w1 << 10) & 0x400) + (w2 >> 6);
    parameterf.m_wC6 = w2 & 0x3f;
    ResetSensor();
    return(true);
}


// Druck und Temperatur auslesen
// Der Druck wird in 1/10 hPa 
// und die Temperatur in 1/10 C zurckgegeben
int DruckTempLesen()
{
int d1, d2, ut1, dut, off, sens, x;

	if (!m_iPort)
	   return(false);

	d1 = ADWertLesen(0x2f);
	d2 = ADWertLesen(0x4f);

	ut1  = 8 * parameterf.m_wC5 + 20224;

	if ((d2 - ut1) > 0)
	   dut = d2 - ut1;
	else
	   dut = (d2-ut1)-(int)((d2-ut1)/128)*((d2-ut1)/128)/4;

	off  = (int)(parameterf.m_wC2*4
		 + (int)((int)((parameterf.m_wC4 - 512) * dut) / 4096));
	sens = parameterf.m_wC1 + ((int)(parameterf.m_wC3 * dut) / 1024) + 24576;
	x    = (int)(sens * (d1 - 7168)) / 16384 - off;
	parameterf.druck = (int)(x * 10 / 32 + 250 * 10);

	if((d2-ut1)>0)
	   parameterf.temp = (int)(200 +
				(int)(dut * (parameterf.m_wC6 + 50)) /1024);
	else
	   parameterf.temp = (int)(200 +
				(int)(dut * (parameterf.m_wC6 + 50)) /1024+dut/256);

	return(true);
}

/**************************************************************************/
/********************** Ringbuffer for the pressure values  ***************/
/**************************************************************************/
/*
 * This fuction was heavily altered by Andreas Theofilu.
 * I added the possibility to read the last values from a file. This
 * fills the ring buffer with the values out of the file.
 */
void init_ring()
{
int i;

	if ((pressure = malloc (RINGLENGTH * sizeof (int))) == NULL)
	{
	   err_sys ("druck: error(init_ring): not enough memory");
	   exit(1);
	}

	for (i = 0; i < RINGLENGTH; i++)
	   pressure[i] = -1;

	first = 0;
	/*
	 * If we have a <data file>, we try to open it. In case it
	 * exists we read it and fill our buffer with it content.
	 * Otherwise we create it and continue to save every value
	 * in it.
	 */
	if (DataFile)
	{
	long pos, anz;
	mpoint mp;
	int flag = 0;
	time_t t, last;
	struct stat st;

	   if ((Datafd = open (DataFile, O_RDWR, 0600)) == -1)
	   {
	      if ((Datafd = open (DataFile, O_RDWR | O_CREAT, 0600)) == -1)
	         err_sys ("druck: error creating a file!");
	   }

	   /*
	    * Here we have an open file. Let's see if it's empty or not.
	    */
	   if (fstat (Datafd, &st) == -1)
	      err_sys ("druck: error stating a file!");

	   anz = st.st_size / sizeof (mpoint);

	   /*
	    * In case "anz" is > 0, we do have something in our file.
	    * Now we must see, if the data is not older than 24 hours.
	    */
	   pos = 0;
	   t = time (NULL);

	   if (anz > RINGLENGTH)
	   {
	      pos = anz - RINGLENGTH;
	      lseek (Datafd, pos * sizeof (mpoint), SEEK_SET);
	   }

	   last = 0;

	   while (pos < anz)
	   {
	      read (Datafd, &mp, sizeof (mpoint));

	      if (last && difftime (last, mp.tstamp) > 140.0)
	      {
	         while (last < mp.tstamp)
		 {
		    pressure[first] = 100;
		    first++;

		    if (first == RINGLENGTH)
		       first = 0;

		    last += 120L;
		 }

		 pos++;
		 last = mp.tstamp;
		 flag = 1;
		 continue;

	      }

	      if (difftime (t, mp.tstamp) < 86400.0)  // 1 day = 86400 seconds
	      {
	         pressure[first] = mp.pressure;
		 parameterf.temp = mp.temperature;
		 first++;

		 if (first >= RINGLENGTH)
		    first = 0;

		 flag = 1;
	         last = mp.tstamp;
	      }

	      pos++;
	   }

	   lseek (Datafd, anz * sizeof(mpoint), SEEK_SET);

	   /*
	    * Because we could have been longer offline as only 2 minutes
	    * we need fo fill the ring buffer with 0's up to the moment.
	    */
	   if (anz > 0 && flag)
	   {
	   int menge;
	   double td;

	      td = difftime (time (NULL), mp.tstamp);
	      menge = (int)(td / 120.0);

	      for (i = 0; i < menge; i++)
	      {
	         pressure[first] = 100;
		 first++;

		 if (first >= RINGLENGTH)
		    first = 0;
	      }
	   }
	}
}

/*
 * This function was altered by Andreas Theofilu.
 * I added the possibility to write the messured values to a
 * file.
 */
void insert_ring (int val)
{
static float erg = -1.0;
static int i = 0;

	if (val <= 0)
	   return;

	// This function is called every 10 seconds. The following
	// condition makes sure, that only every 2 minutes something
	// goes to the ring buffer and to a file, if we have one.
	if (++i == 12)
	{
	   pressure[first] = (int) erg;

	   // Write this also to a file, if we have one.
	   if (DataFile)
	   {
	   mpoint mp;

	      mp.pressure = (int)erg;
	      mp.temperature = parameterf.temp;
	      mp.tstamp = time (NULL);

	      write (Datafd, &mp, sizeof (mpoint));
	   }

	   if (++first == RINGLENGTH)
	      first = 0;

	   i = 0;
	}

	if (erg <= 0)
	   erg = (float) val;
	else
	   erg = 0.9 * erg + 0.1 * (float) val;
}

/**************************************************************************/
/***************************** The two threads  ****************************/
/**************************************************************************/

void *pthr_getpressure( void *pV_data )
{
time_t now, t;

	t = time (NULL);

	while (1)
	{
	   pthread_mutex_lock(&fastmutex);
	   now = time(NULL);
	   DruckTempLesen();
	   insert_ring (parameterf.druck);
//	   printf("%d %d %s", parameterf.druck, parameterf.temp, ctime(&now));
//	   fflush(stdout);
	   pthread_mutex_unlock(&fastmutex);
	   sleep(1);

	   // sleep for 10 seconds
	   while (difftime ((now = time(NULL)), t) < 10)
	      usleep (100 * 1000);

	   t = now;
	}

	return NULL;
}


// including socket

void *pthr_sendpressure( void *pV_data )
{
int i;
static int count=0;
char *buf;
char *bp;
// socket structure
struct sockaddr_in  client1, server1;
struct servent *serviceInfo;
//struct hostent *hostinfo;
int s1,s;
socklen_t length;

	// socket server
	if ((serviceInfo = getservbyname ("druck", "tcp")))
	   server1.sin_port = serviceInfo->s_port;

	server1.sin_family = AF_INET;
	server1.sin_addr.s_addr = INADDR_ANY;
	s = socket(AF_INET,SOCK_STREAM,0);

	if (s < 0)
	   err_sys ("druck: Error in socket");

	if (bind (s, (struct sockaddr *)&server1, sizeof (server1)) < 0)
	   err_sys ("druck: Error in bind");

	if (listen (s, 5) < 0)
	   err_sys ("druck: Error in listen");

	length = sizeof(client1);

	// transferbuffer
	if ((buf = malloc (RINGLENGTH * 5)) == NULL)
	   err_sys ("druck: error (sendpressure): not enough memory.");

        syslog (LOG_INFO, "druck: Server ready: %d",s);

	while (1)
	{
	   if ((s1 = accept (s, (struct sockaddr *)&client1, &length)) < 0)
	      err_sys ("druck: Error in accept: %d", s1);

	   pthread_mutex_lock (&fastmutex);
	   bp = buf;
	   count = first;  // first stands of i+1

	   if (count < 0)
	      count = RINGLENGTH - 1;

	   for (i = 0; i < RINGLENGTH; i++)
	   {
	      if (count + i >= RINGLENGTH)
	         count = -i;

	      sprintf (bp, "%5d", pressure[count+i]);
	      bp += 5;
	   }

	   syslog (LOG_INFO, "druck: Connected to client");
	   write (s1, buf, 5 * RINGLENGTH);
	   pthread_mutex_unlock(&fastmutex);
	   close (s1);
	}

	return NULL;
}

/*
 * Inserted by Andreas Theofilu (Theos Soft) <andreas@TheosSoft.net>.
 * This function makes the program to detach from console.
 */
void daemon_start (int ignsigcld)
{
int childpid, fd;

        if (getpid () == 1)
           goto out;

#ifdef SIGTTOU
        signal (SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
        signal (SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
        signal (SIGTSTP, SIG_IGN);
#endif

        if ((childpid = fork ()) < 0)
           err_sys ("Can't fork this child");
        else if (childpid > 0)
           exit (0);

#if defined(SIGTSTP) && !defined(sinix) && !defined(Linux)      /* BSD */
        if (setpgrp (0, getpid ()) == -1)
           err_sys ("Can't change process group");

        if ((fd = open ("/dev/tty", O_RDWR)) >= 0)
        {
           ioctl (fd, TIOCNOTTY, (char *)NULL);
           close (fd);
        }
#else                   /* System V */
        if (setpgrp () == -1)
           err_sys ("Can't change process group");

        signal (SIGHUP, SIG_IGN);

        if ((childpid = fork ()) < 0)
           err_sys ("Can't fork second child");
        else if (childpid > 0)
           exit (0);            /* first child */

        /* second child */
#endif

out:
        for (fd = 0; fd < NOFILE; fd++)
           close (fd);

        errno = 0;
        chdir ("/");
        umask (0);

        if (ignsigcld)
        {
#ifdef SIGTSTP
           signal (SIGCLD, sig_child);
#else
           signal (SIGCLD, SIG_IGN);
#endif
        }
}

void sig_child ()
{
#if defined(BSD) && !defined(sinix)
int pid;
union wait status;

        while ((pid = wait3 (&status, WNOHANG, (struct rusage *)0)) > 0)
                ;
#endif
}

/*
 * The MAIN function. Altered by Andreas Theofilu.
 * I added checks for 2 command line parameters. However. It's still
 * possible to call the daemon the old way, with only one parameter
 * that defines the parallel port with the sensor.
 */
int main(int argc, char **argv)
{
char *pC_port;

	/*
	 * First we initialize our new daemon and detaching from
	 * the console. Everey message goes now to the systems
	 * logs!
	 */
	pname = argv[0];
	daemon_start (0);

	if ( argc < 2 )
	   err_quit ( "usage: %s -p <LPT1|LPT2> [-b <datafile>]", pname);

	/*
	 * Parsing the command line...
	 */
	argc--;
	argv++;
	pC_port = DataFile = NULL;

	while (argc)
	{
	   if (strncmp (*argv, "-p", 2) == 0)
	   {
	      argc--;
	      argv++;

	      if (!argc)
	      {
	         syslog (LOG_ERR, "%s: Parameter \"-p\" needs a port (LPT1|LPT2)!", pname);
	   	 err_quit ( "usage: %s -p <LPT1|LPT2> [-b <datafile>]", pname);
	      }


	      pC_port    = *argv;
	   }

	   if (strncmp (*argv, "-b", 2) == 0)
	   {
	      argc--;
	      argv++;

	      if (!argc)
	      {
	         syslog (LOG_ERR, "%s: Parameter \"-b\" needs a filename as parameter!", pname);
	   	 err_quit ( "usage: %s -p <LPT1|LPT2> [-b <datafile>]", pname);
	      }

	      DataFile = *argv;
	   }

	   argc--;
	   argv++;
	}

	if (pC_port && strncmp( pC_port, "LPT2", 4 ) == 0 )
	   m_iPort = D_LPT2_DATA;
	else
	   m_iPort = D_LPT1_DATA;

	/*### Abschnitt 3 */
	/*
	 * The following call works only on Linux and requires root privileges!
	 * It sets the port access permission bits for the process.
	 */
	if  (ioperm (m_iPort, 3, 1) != 0)	/* get 10 perm */
	   err_quit ( "ERROR: can_t open LPT device for read/write!" );

	/* ### Abschnitt 4 */
/*	printf( "\nPress ^C to stop program.\n"); */
	init_ring();
	ParameterLesen();

	if ( pthread_create(&pthr_get, NULL, pthr_getpressure, (void *)0) != 0)
	   err_sys ( "Create of task 1 failed");

	if ( pthread_create(&pthr_send, NULL, pthr_sendpressure,(void *)1) != 0)
	   err_sys ( "Create of task 2 failed");

	while ( 1 == 1 )
	{
	   syslog (LOG_INFO, "%s: I'm alive ...", pname);
	   sleep( 3600 );
	}

	return( 0 );
}

