#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>

#define TNONE		0
#define TMONTH		1
#define TYEAR		2

#define MONTH		2592000
#define YEAR		31104000

#define HIGH_NN 256.0      // 35m over NN
/*
 * Some global variables we need in this module.
 */
char *DataFile;
int Datafd;
int trange = TNONE;

typedef struct
{
	int pressure;
	int temperature;
	time_t tstamp;
} mpoint;

void usage ();

int main (int argc, char *argv[])
{
time_t t, lt;
struct tm *zeit;
mpoint mp;
struct stat st;
long anz, j, a;
unsigned long tr = MONTH;
double delta_p, press, opress;

	argc--;
	*argv++;

	if (argc <= 0)
	{
	   fprintf (stderr, "No parameters found on command line!\n");
	   usage ();
	   return 1;
	}

	if (!strcasecmp ("-m", *argv))
	{
	   trange = TMONTH;
	   argc--;
	   *argv++;
	}

	if (!strcasecmp ("-y", *argv))
	{
	   trange = TYEAR;
	   argc--;
	   *argv++;
	}

	if (strlen (*argv) < 1)
	{
	   fprintf (stderr, "No file name to read data from was given!\n");
	   usage ();
	   return 1;
	}
	else
	   DataFile = (char *)strdup (*argv);

	if (trange == TNONE)
	{
	   fprintf (stderr, "You have not set the time range to be showed!\n");
	   free (DataFile);
	   usage ();
	   return 1;
	}

	/*
	 * First me must get the actual date (and time) to find out what
	 * month and/or year we should show.
	 * We always read back a month or a year from the actual date.
	 */
	t = time (NULL);
	zeit = gmtime (&t);
	zeit->tm_hour = 0;
	zeit->tm_min = 0;
	zeit->tm_sec = 0;
	t = mktime (zeit);

	/*
	 * Oeffnen der angegebenen Datei und lesen der Daten.
	 */
	if ((Datafd = open (DataFile, O_RDONLY, 0400)) == -1)
	{
	   fprintf (stderr, "Illegal file name \"%s\" or no permission!\n", DataFile);
	   free (DataFile);
	   usage ();
	   return 1;
	}

	if (fstat (Datafd, &st) == -1)
	{
	   fprintf (stderr, "Error stating file \"%s\"!\n", DataFile);
	   free (DataFile);
	   close (Datafd);
	   exit (1);
	}

	anz = st.st_size / sizeof (mpoint);

	if (anz <= 0L)
	{
	   fprintf (stderr, "File is empty!\n");
	   free (DataFile);
	   close (Datafd);
	   exit (1);
	}

	/*
	 * Nun suchen wir den Zeitpunkt ab dem wir die Daten
	 * augeben wollen und starten die Datenausgabe.
	 */
	switch (trange)
	{
	   case TMONTH:
	      tr = MONTH;
	   break;

	   case TYEAR:
	      tr = YEAR;
	   break;
	}

	j = a = 0L;
	lt = 0L;
	press = opress = 0.0;
	delta_p = pow ((1.0 - (6.5 * HIGH_NN) / 288000.0), 5.255);

	while (j < anz)
	{
	   read (Datafd, &mp, sizeof (mpoint));
	   press = ((double)mp.pressure / 10.0) / delta_p;

	   if (opress != 0.0 && ((opress - press) > 8 || (press - opress) > 8))
	      press = opress;

	   if (mp.tstamp >= (t - tr))
	   {
	      if (lt && (mp.tstamp - lt) > 130L)
	      {
	         while (lt < mp.tstamp)
		 {
		    printf ("%f %d\n", (double)a / 720.0, (int)press);
		    lt += 120L;
		    a++;
		 }
	      }

	      printf ("%f %d\n", (double)a / 720.0, (int)press);
	      lt = mp.tstamp;
	      a++;
	   }

	   if (mp.tstamp >= t)
	      break;

	   j++;
	   opress = press;
	}

	close (Datafd);
	free (DataFile);
	return 0;
}

void usage ()
{
	printf ("Version 1.0\n");
	printf ("Author: Andreas Theofilu <andreas@TheosSoft.net>\n");
	printf ("License: GNU General Public License\n\n");
	printf ("readdruck <-m | -y> <file>\n\n");
	printf ("   -m     Read the last 30 days from <file>\n\n");
	printf ("   -y     Read the last 360 days from <file>\n\n");
	printf (" <file>   The file where the messured data are saved. This\n");
	printf ("          is produced by <druckd>, the daemon who reads\n");
	printf ("          the air pressure sensor.\n");
}

